//
//  OTRLibrary.m
//  OldTimeRadio
//
//  Created by Brian Moakley on 6/1/14.
//  Copyright (c) 2014 Razeware. All rights reserved.
//

#import <sqlite3.h>
#import "OTRLibrary.h"
#import "SQLLiteManager.h"
#import "OTRShow.h"
#import "OTREpisode.h"
#import "OTRFavorite.h"
#import "OTRSite.h"

@interface OTRLibrary()

@property (strong, nonatomic) NSMutableArray * shows;
@property (strong, nonatomic) NSMutableArray * episodes;
@property (strong, nonatomic) NSMutableArray * favorites;
@property (strong, nonatomic) NSMutableArray * sites;

@end

@implementation OTRLibrary

- (id)init
{
    self = [super init];
    if (self)
    {
        _episodes = [[NSMutableArray alloc] init];
        _shows = [[NSMutableArray alloc] init];
        _sites = [[NSMutableArray alloc] init];
        _favorites = [[NSMutableArray alloc] init];
    }
    return self;
}

+ (instancetype)sharedLibrary {
    static OTRLibrary *sharedMyManager = nil;
    @synchronized(self) {
        if (sharedMyManager == nil)
        {
            sharedMyManager = [[self alloc] init];
        }
    }
    return sharedMyManager;
}

-(void) initializeLibrary
{
    SQLLiteManager * sqlite = [SQLLiteManager sharedManager];
    [sqlite open];
    
    NSString * sql = @"select * from Shows";
    sqlite3_stmt * selectStatement;
    if (sqlite3_prepare_v2(sqlite.database, [sql UTF8String], -1, &selectStatement, NULL) == SQLITE_OK)
    {
        while (sqlite3_step(selectStatement) == SQLITE_ROW)
        {
            int showID = sqlite3_column_int(selectStatement, 0);
            char * title = (char *) sqlite3_column_text(selectStatement, 1);
            char * fileName = (char *) sqlite3_column_text(selectStatement, 2);
            char * note = (char *) sqlite3_column_text(selectStatement, 3);
            
            OTRShow * show = [[OTRShow alloc] init];
            show.title = [NSString stringWithUTF8String:title];
            show.thumbnailFileName = [NSString stringWithUTF8String:fileName];
            show.showDescription = [NSString stringWithUTF8String:note];
            show.showId = showID;
            
            [self.shows addObject:show];
        }
    }
    
    sql = @"select * from Episodes";
    selectStatement = NULL;
    if (sqlite3_prepare_v2(sqlite.database, [sql UTF8String], -1, &selectStatement, NULL) == SQLITE_OK)
    {
        while (sqlite3_step(selectStatement) == SQLITE_ROW)
        {
            int episodeId = sqlite3_column_int(selectStatement, 0);
            int showId = sqlite3_column_int(selectStatement, 1);
            char * title = (char *) sqlite3_column_text(selectStatement, 2);
            char * broadcastDate = (char *) sqlite3_column_text(selectStatement, 3);
            char * fileName = (char *) sqlite3_column_text(selectStatement, 4);
            
            NSDateFormatter * dataFormatter = [[NSDateFormatter alloc] init];
            [dataFormatter setDateFormat:@"yyyy-MM-dd HH:mm:ss +0000"];
            NSString * broadcastString = [NSString stringWithUTF8String:broadcastDate];
            NSDate * date = [dataFormatter dateFromString:broadcastString];
            
            OTREpisode * episode = [[OTREpisode alloc] init];
            episode.episodeId = episodeId;
            episode.title = [NSString stringWithUTF8String:title];
            episode.fileLocation = [NSString stringWithUTF8String:fileName];
            episode.broadcastDate = date;
            
            for (OTRShow * show in self.shows)
            {
                if (show.showId == showId)
                {
                    episode.parentShow = show;
                }
            }
            
            [self.episodes addObject:episode];
        }
    }
    
    sql = @"select * from Favorites";
    if (sqlite3_prepare_v2(sqlite.database, [sql UTF8String], -1, &selectStatement, NULL) == SQLITE_OK)
    {
        while (sqlite3_step(selectStatement) == SQLITE_ROW)
        {
            int episodeId = sqlite3_column_int(selectStatement, 1);
            char * note = (char *) sqlite3_column_text(selectStatement, 2);
            char * favoriteDate = (char *) sqlite3_column_text(selectStatement, 3);
            
            NSDateFormatter * dataFormatter = [[NSDateFormatter alloc] init];
            [dataFormatter setDateFormat:@"yyyy-MM-dd HH:mm:ss +0000"];
            NSString * favoriteString = [NSString stringWithUTF8String:favoriteDate];
            NSDate * date = [dataFormatter dateFromString:favoriteString];

            OTRFavorite * favorite = [[OTRFavorite alloc] init];
            favorite.note = [NSString stringWithUTF8String:note];
            favorite.favoriteDate = date;
            
            for (OTREpisode * episode in self.episodes)
            {
                if (episode.episodeId == episodeId)
                {
                    favorite.episode = episode;
                }
            }
            
            [self.favorites addObject:favorite];
        }
    }
    
    sql = @"select * from Sites";
    if (sqlite3_prepare_v2(sqlite.database, [sql UTF8String], -1, &selectStatement, NULL) == SQLITE_OK)
    {
        while (sqlite3_step(selectStatement) == SQLITE_ROW)
        {
            char * name = (char *) sqlite3_column_text(selectStatement, 1);
            char * address = (char *) sqlite3_column_text(selectStatement, 2);
            int isFree = sqlite3_column_int(selectStatement, 3);
            int isMembershipRequired = sqlite3_column_int(selectStatement, 4);
        
            OTRSite * site = [[OTRSite alloc] init];
            site.name = [NSString stringWithUTF8String:name];
            site.address = [NSString stringWithUTF8String:address];
            site.isFree = [@(isFree) boolValue];
            site.isMembershipRequired = [@(isMembershipRequired) boolValue];

            [self.sites addObject:site];
        }
    }
    
    
    [sqlite close];
}

-(void) buildLibraryFromPlist
{
    SQLLiteManager * sqlLite = [SQLLiteManager sharedManager];
    
    if ([sqlLite open] == SQLITE_OK)
    {
        
        NSArray * sqlStatements = @[ @"CREATE TABLE if not exists Sites (SiteID integer primary key autoincrement, Name text, Address text, IsFree BOOLEAN, IsMembershipRequired BOOLEAN);",
                                     @"CREATE TABLE if not exists Shows (ShowID integer primary key autoincrement, Title text, Thumbnail text, Description text);",
                                     @"CREATE TABLE if not exists Episodes (EpisodeID integer primary key autoincrement, ShowID int, Title text, BroadcastDate date, FileLocation text);",
                                     @"CREATE TABLE if not exists Favorites (FavoriteID integer primary key autoincrement, EpisodeID integer, Note text, FavoriteDate date);"
                                     ];
        
        for (NSString * sqlStatement in sqlStatements)
        {
            const char * sql = [sqlStatement UTF8String];
            sqlite3_exec(sqlLite.database, sql, NULL, NULL, NULL);
        }
        NSString * plistPath = [[NSBundle mainBundle] pathForResource:@"old_time_radio" ofType:@"plist"];
        NSURL * plistURL = [NSURL fileURLWithPath:plistPath];
        NSData * data = [NSData dataWithContentsOfURL:plistURL options:NSDataReadingMappedIfSafe error:nil];
        
        NSDictionary * plist = [NSPropertyListSerialization propertyListFromData:data mutabilityOption:NSPropertyListImmutable format:nil errorDescription:nil];
        NSArray * sites = plist[@"Sites"];
        for (NSDictionary * site in sites)
        {
            NSNumber * isFree = site[@"Is Free"];
            NSNumber * isMembershipRequired = site[@"Membership Required"];
            NSString * insertSql = [NSString stringWithFormat:@"insert into Sites (Name, Address, IsFree, IsMembershipRequired) values ('%@', '%@', %d, %d);", site[@"Name"], site[@"Address"], isFree.intValue, isMembershipRequired.intValue];
            sqlite3_stmt * statement = NULL;
            const char * sql = [insertSql UTF8String];
            sqlite3_prepare_v2(sqlLite.database, sql, -1, &statement, NULL);
            if (sqlite3_step(statement) == SQLITE_DONE)
            {
                NSLog(@"Site Inserted");
            }
            sqlite3_finalize(statement);
        }
        NSArray * favorites = plist[@"Favorites"];
        for (NSDictionary * favorite in favorites)
        {
            NSNumber * episodeId = favorite[@"Episode Id"];
            NSString * insertSql = [NSString stringWithFormat:@"insert into Favorites (EpisodeID, Note, FavoriteDate) values (%d, '%@', '%@');", episodeId.intValue, [favorite[@"Note"] stringByReplacingOccurrencesOfString:@"'" withString:@"''"], favorite[@"Favorite Date"]];
            
            sqlite3_stmt * statement = NULL;
            const char * sql = [insertSql UTF8String];
            
            sqlite3_prepare_v2(sqlLite.database, sql, -1, &statement, NULL);
            if (sqlite3_step(statement) == SQLITE_DONE)
            {
                NSLog(@"Favorite Inserted");
            }
            sqlite3_finalize(statement);
        }
        NSArray * episodes = plist[@"Episodes"];
        for (NSDictionary * episode in episodes)
        {
            NSNumber * showId = episode[@"Show Id"];
            NSString * insertSql = [NSString stringWithFormat:@"insert into Episodes (ShowID, Title, BroadcastDate, FileLocation) values (%d, '%@', '%@', '%@');", showId.intValue, [episode[@"Title"] stringByReplacingOccurrencesOfString:@"'" withString:@"''"], episode[@"Broadcast Date"], episode[@"File Location"]];
            
            sqlite3_stmt * statement = NULL;
            const char * sql = [insertSql UTF8String];
            
            sqlite3_prepare_v2(sqlLite.database, sql, -1, &statement, NULL);
            if (sqlite3_step(statement) == SQLITE_DONE)
            {
                NSLog(@"Episode Inserted");
            }
            sqlite3_finalize(statement);
        }
        NSArray * shows = plist[@"Shows"];
        for (NSDictionary * show in shows)
        {
            NSString * insertSql = [NSString stringWithFormat:@"insert into Shows (Title, Thumbnail, Description) values ('%@', '%@', '%@');", [show[@"Title"] stringByReplacingOccurrencesOfString:@"'" withString:@"''"], show[@"Thumbnail"], [show[@"Description"] stringByReplacingOccurrencesOfString:@"'" withString:@"''"]];
            
            sqlite3_stmt * statement = NULL;
            const char * sql = [insertSql UTF8String];
            
            sqlite3_prepare_v2(sqlLite.database, sql, -1, &statement, NULL);
            if (sqlite3_step(statement) == SQLITE_DONE)
            {
                NSLog(@"Show Inserted");
            }
            sqlite3_finalize(statement);
        }
    }
    
    [sqlLite close];
}

@end
